import numpy as np
import torch
import torch_geometric
import random
from sklearn.cluster import AffinityPropagation
import pickle
import os
from torch_geometric.data import TemporalData

from perturb_fidelity import get_datasets, save_dataset


def trim_ds(dataset, node_to_mode, mode_clusters, num_events=2000):

    trimmed_dataset = TemporalData(
        src=dataset.src[:num_events],
        dst=dataset.dst[:num_events],
        t=dataset.t[:num_events],
        msg=dataset.msg[:num_events] if dataset.msg is not None else None,
    )

    unique_nodes = set(trimmed_dataset.src.tolist()) | set(trimmed_dataset.dst.tolist())  # type: ignore

    trimmed_node_to_mode = {
        node: mode for node, mode in node_to_mode.items() if node in unique_nodes
    }

    unique_modes = set(trimmed_node_to_mode.values())

    trimmed_mode_clusters = mode_clusters[list(unique_modes)]

    return trimmed_dataset, trimmed_node_to_mode, trimmed_mode_clusters


def get_modes(node_embeddings, dataset_name, cache_dir="modes_cache"):

    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    cache_file = os.path.join(cache_dir, f"{dataset_name}_modes.pkl")

    if os.path.exists(cache_file):
        print(f"Loading cached modes from {cache_file}...")
        with open(cache_file, "rb") as f:
            cached_data = pickle.load(f)
            return cached_data["node_to_mode"], cached_data["mode_clusters"]

    # Convert the node embeddings to a numpy array
    if isinstance(node_embeddings, torch.Tensor):
        node_embeddings = node_embeddings.cpu().detach().numpy()

    clustering = AffinityPropagation(random_state=5).fit(node_embeddings)

    cluster_labels = clustering.labels_

    # Create a mapping of node ID to assigned mode
    node_to_mode = {
        node_id: cluster_labels[node_id] for node_id in range(len(cluster_labels))
    }

    mode_clusters = clustering.cluster_centers_

    with open(cache_file, "wb") as f:
        pickle.dump({"node_to_mode": node_to_mode, "mode_clusters": mode_clusters}, f)

    print(f"Modes cached to {cache_file}.")

    return node_to_mode, mode_clusters


def mode_dropping(data, node_modes, cluster_centers, p):

    # Identify modes to drop with probability p
    num_modes = len(cluster_centers)
    modes_to_drop = [mode for mode in range(num_modes) if random.random() < p]

    # Identify safe events (events that don't involve nodes from dropped modes)
    safe_indices = []
    dropped_indices = []

    src = data.src.clone()
    dst = data.dst.clone()
    msg = data.msg.clone()

    for i in range(len(src)):
        src_node = src[i].item()
        dst_node = dst[i].item()

        src_mode = node_modes[src_node]
        dst_mode = node_modes[dst_node]

        if src_mode in modes_to_drop or dst_mode in modes_to_drop:
            dropped_indices.append(i)
        else:
            safe_indices.append(i)

    if not safe_indices:
        print("Warning: No safe events found. Using random event replication.")
        safe_indices = list(
            range(len(src))
        )  # If no safe events, fallback to entire dataset as safe

    # Replace events that involve dropped modes with random safe events
    for idx in dropped_indices:
        replacement_idx = random.choice(safe_indices)

        src[idx] = src[replacement_idx]
        dst[idx] = dst[replacement_idx]
        msg[idx] = msg[replacement_idx]

    modified_data = TemporalData(src=src, dst=dst, t=data.t, msg=msg)

    return modified_data


def mode_collapse(data, node_modes, cluster_centers, p):

    if p == 0.0:
        return data

    src = data.src.clone()
    dst = data.dst.clone()
    msg = data.msg.clone()

    # Calculate representative message for each mode
    num_modes = len(cluster_centers)
    mode_messages = {mode: [] for mode in range(num_modes)}

    # Group messages by node's mode
    for i in range(len(src)):
        src_node = src[i].item()
        dst_node = dst[i].item()

        # Append messages
        src_mode = node_modes[src_node]
        dst_mode = node_modes[dst_node]

        mode_messages[src_mode].append(msg[i].cpu().numpy())
        mode_messages[dst_mode].append(msg[i].cpu().numpy())

    representative_messages = {}
    for mode, messages in mode_messages.items():
        if len(messages) > 0:
            representative_messages[mode] = np.mean(messages, axis=0)
        else:
            representative_messages[mode] = np.zeros(msg.shape[1])

    for mode in representative_messages:
        representative_messages[mode] = torch.tensor(
            representative_messages[mode], dtype=msg.dtype
        )

    for i in range(len(src)):
        if random.random() < p:
            src_node = src[i].item()
            dst_node = dst[i].item()

            src_mode = node_modes[src_node]
            dst_mode = node_modes[dst_node]

            src[i] = torch.tensor(
                np.argmax(
                    np.linalg.norm(cluster_centers - cluster_centers[src_mode], axis=1)
                )
            )
            dst[i] = torch.tensor(
                np.argmax(
                    np.linalg.norm(cluster_centers - cluster_centers[dst_mode], axis=1)
                )
            )

            msg[i] = representative_messages[src_mode]

    collapsed_data = TemporalData(src=src, dst=dst, t=data.t, msg=msg)

    return collapsed_data


def apply_perturbations_diversity(
    data, dataset_name, node_modes, cluster_centers, seeds=10, ps=None, output_dir=None
):

    if ps is None:
        ps = np.linspace(0, 0.9, 10)

    for seed in range(seeds):
        random.seed(seed)
        torch_geometric.seed_everything(seed)
        for p in ps:

            perturbed_data = mode_dropping(data, node_modes, cluster_centers, p)
            save_dataset(
                perturbed_data, dataset_name, "mode_droping", seed, p, output_dir
            )

            perturbed_data = mode_collapse(data, node_modes, cluster_centers, p)
            save_dataset(
                perturbed_data, dataset_name, "mode_collapse", seed, p, output_dir
            )


if __name__ == "__main__":

    output_dir = "./DATA_DIR"
    os.makedirs(output_dir, exist_ok=True)

    NUM_EVENTS = 1000

    ROOT_MEM_DIR = "./tgn"
    datasets = get_datasets(grid_nn=10000, grid_max_time=10000, grid_seed=101)

    dataset_names = ["Reddit", "Wikipedia", "MOOC", "LastFM", "Grid"]
    num_seeds = 5

    node_modes_list = []
    cluster_center_list = []
    for name in dataset_names:
        print(f"Processing {name}")
        memory = torch.load(f"{ROOT_MEM_DIR}/{name}.pt")
        node_embedding = memory.memory
        modes, clusters_centers = get_modes(node_embedding, name)
        node_modes_list.append(modes)
        cluster_center_list.append(clusters_centers)

    for dataset, name, node_modes, cluster_centers in zip(
        datasets, dataset_names, node_modes_list, cluster_center_list
    ):
        dataset.y = None  # type: ignore

        dataset, _, _ = trim_ds(
            dataset,
            node_modes,
            cluster_centers,
            num_events=NUM_EVENTS,
        )
        print(f"Applying perturbations for dataset: {name}")
        apply_perturbations_diversity(
            dataset,
            name,
            node_modes,
            cluster_centers,
            seeds=num_seeds,
            output_dir=output_dir,
        )
